import pytest
from datetime import datetime, timedelta


def test_standard_values(ini_config):
    assert ini_config.one['string_exp'] == 'testerstr'
    assert ini_config.one['int_exp'] == 6
    assert ini_config.one['bool_exp'] is True
    assert ini_config.one['multiword_string'] == 'this is a test of a sentence'
    assert ini_config.one['multiword_comma_string'] == 'this is a test of a sentence with an, comma in it'


def test_list_values(ini_config):
    assert ini_config.two['list_str_exp'] == ["first", "second", "third"]
    assert ini_config.two['list_int_exp'] == [1, 2, 3]
    assert ini_config.two['list_bool_exp'] == [True, False, True, False, False, True]
    assert ini_config.two['list_mixed_exp'] == ["big", 12, "small", False, True, 69]
    assert ini_config.two['multiword_string_list'] == ["this is a", "test of a sentence", "that is split inside a list"]


def test_list_bad_format(ini_config):
    assert ini_config.three['list_bad_format'] == ["first", "second", "third", "fourth"]


def test_datetime_values(ini_config):
    assert ini_config.four['exp_datetime'] == datetime(year=2020, month=11, day=13, hour=1, minute=55, second=33)
    assert ini_config.four['exp_timedelta'] == timedelta(days=3, hours=1, minutes=21, seconds=7)
    assert ini_config.four['exp_timedelta_neg'] == timedelta(days=-10, hours=-6, minutes=-7, seconds=-8)


def test_add_options(ini_config, sample_ini_file):
    _string_test = {'new_string_key': 'new_string_value'}
    ini_config.set_one(**_string_test)
    assert ini_config.one['new_string_key'] == 'new_string_value'

    _bool_test = {'new_bool_test': True}
    ini_config.set_one(**_bool_test)
    assert ini_config.one['new_bool_test'] is True

    _integer_test = {'new_integer_test': 23}
    ini_config.set_one(**_integer_test)
    assert ini_config.one['new_integer_test'] == 23

    _list_test = {'new_list_value': [1, 2, 'alpha', 'Beta', True, False]}
    ini_config.set_two(**_list_test)
    assert ini_config.two['new_list_value'] == [1, 2, 'alpha', 'Beta', True, False]

    with open(sample_ini_file, 'r') as sample_file:
        _content = sample_file.read()
    for line in _content.splitlines():
        if '=' in line:
            _key, _value = line.split('=')
            _key = _key.strip()
            _value = _value.strip()
            if _key == 'new_string_key':
                assert _value == 'new_string_value'
            elif _key == 'new_list_value':
                assert _value == '-LIST- 1, 2, alpha, Beta, True, False'
            elif _key == 'new_bool_test':
                assert _value == 'True'
            elif _key == 'new_integer_test':
                assert _value == '23'


def test_new_section(ini_config, sample_ini_file):
    with pytest.raises(AttributeError):
        a = ini_config.five

    ini_config.new_section('five')
    assert ini_config.five == {}

    ini_config.new_section('six', sixer_int=5)
    assert ini_config.six == {'sixer_int': 5}

    with open(sample_ini_file, 'r') as sample_file:
        _content = sample_file.read()
    _section_list = [line.strip() for line in _content.splitlines() if '=' not in line]

    assert '[five]' in _section_list
    assert '[six]' in _section_list




# region [Imports]

# * Standard Library Imports -->
import os
import configparser
from typing import Union
from datetime import datetime, timedelta

# * Third Party Imports -->
from fuzzywuzzy import fuzz

# * Gid Imports -->
import gidlogger as glog

# * Local Imports -->
from gidconfig.data.enums import Get

# endregion [Imports]


# region [Logging]

log = glog.aux_logger(__name__)
log.info(glog.imported(__name__))

# endregion [Logging]


class ConfigHandler(configparser.ConfigParser):
    def __init__(self, config_file=None, auto_read=True, auto_save=True, allow_no_value=True, **kwargs):
        super().__init__(**kwargs, allow_no_value=allow_no_value)
        self.config_file = '' if config_file is None else config_file
        self.auto_read = auto_read
        self.auto_save = auto_save
        self._method_select = {Get.basic: self.get, Get.boolean: self.getboolean, Get.int: self.getint, Get.list: self.getlist, Get.path: self.get_path, Get.datetime: self.get_datetime}

        if self.auto_read is True:
            self.read(self.config_file)

    def getlist(self, section, key, delimiter=',', as_set=False):
        _raw = self.get(section, key).strip()
        if _raw.endswith(delimiter):
            _raw = _raw.rstrip(delimiter)
        if _raw.startswith(delimiter):
            _raw = _raw.lstrip(delimiter).strip()
        _out = _raw.replace(delimiter + ' ', delimiter).split(delimiter)
        _out = list(map(lambda x: x.strip(), _out))
        if as_set is True:
            _out = set(_out)
        return _out

    def list_from_keys_only(self, section, as_set=True):
        _result = self.options(section)
        _out = []
        for line in _result:
            if line != '':
                _out.append(line)
        if as_set is True:
            _out = set(_out)
        return _out

    def get_path(self, section, key, cwd_symbol='+cwd+'):
        _raw_path = self.get(section, key)
        if cwd_symbol in _raw_path:
            _out = _raw_path.replace(cwd_symbol, os.getcwd()).replace('\\', '/')
        elif '+userdata+' in _raw_path:
            _out = _raw_path.replace('+userdata+', os.getenv('APPDATA')).replace('\\', '/')
        elif _raw_path == 'notset':
            _out = None
        else:
            _out = os.path.join(_raw_path).replace('\\', '/')
        return _out

    def _best_fuzzymatch(self, in_term, in_targets: Union[list, set, frozenset, tuple, dict]):
        # Todo: replace with process.extractOne() from fuzzywuzzy!
        _rating_list = []
        for _target in in_targets:
            _rating_list.append((_target, fuzz.ratio(in_term, _target)))
        _rating_list.sort(key=lambda x: x[1], reverse=True)
        log.debug("with a fuzzymatch, the term '%s' was best matched to '%s' with and Levenstein-distance of %s", in_term, _rating_list[0][0], _rating_list[0][1])
        return _rating_list[0][0]

    def get_timedelta(self, section, key, amount_seperator=' ', delta_seperator=','):
        _raw_timedelta = self.get(section, key)
        if _raw_timedelta != 'notset':
            _raw_timedelta_list = _raw_timedelta.split(delta_seperator)
            _arg_dict = {'days': 0,
                         'seconds': 0,
                         'microseconds': 0,
                         'milliseconds': 0,
                         'minutes': 0,
                         'hours': 0,
                         'weeks': 0}
            for raw_delta_data in _raw_timedelta_list:
                _amount, _typus = raw_delta_data.strip().split(amount_seperator)
                _key = self._best_fuzzymatch(_typus, _arg_dict)
                _arg_dict[_key] = float(_amount) if '.' in _amount else int(_amount)
            return timedelta(**_arg_dict)

    def get_datetime(self, section, key, dtformat=None):
        _date_time_string = self.get(section, key)
        if _date_time_string == "notset":
            return None
        _dtformat = '%Y-%m-%d %H:%M:%S' if dtformat is None else dtformat
        return datetime.strptime(_date_time_string, _dtformat).astimezone()

    def set_datetime(self, section, key, datetime_object, dtformat=None):
        _dtformat = '%Y-%m-%d %H:%M:%S' if dtformat is None else format
        self.set(section, key, datetime_object.strftime(_dtformat))
        if self.auto_save is True:
            self.save()

    def set(self, section, option, value):
        if isinstance(value, list):
            value = ', '.join(list(map(str, value)))
        if not isinstance(value, str):
            value = str(value)
        super().set(section, option, value)
        if self.auto_save is True:
            self.save()

    def enum_get(self, section: str, option: str, typus: Get = Get.basic):
        return self._method_select.get(typus, self.get)(section, option)

    def save(self):
        with open(self.config_file, 'w') as confile:
            self.write(confile)
        self.read()

    def read(self, filenames=None):
        _configfile = self.config_file if filenames is None else filenames

        super().read(self.config_file)


if __name__ == '__main__':
    pass



import pytest
from gidconfig.experimental.experimental_ini import GidAttConfigIni
import os

SAMPLE_INI_CONTENT = """[one]
string_exp = testerstr
int_exp = 6
bool_exp = yes
multiword_string = this is a test of a sentence
multiword_comma_string = this is a test of a sentence with an, comma in it

[two]
list_str_exp = -LIST- first, second, third
list_int_exp = -LIST- 1, 2, 3
list_bool_exp = -LIST- yes, no, True, false, off, on
list_mixed_exp = -LIST- big, 12, small, off, true, 69
multiword_string_list = -LIST- this is a, test of a sentence, that is split inside a list

[three]
list_bad_format = -LIST- first,second, third,     fourth

[four]
exp_datetime = -DATETIME- 2020-11-13T01:55:33
exp_timedelta = -TIMEDELTA- days: 3, hours: 1, minutes: 21, seconds: 7
exp_timedelta_neg = -TIMEDELTA- negative days: 10, hours: 6, minutes: 7, seconds: 8
"""


@pytest.fixture
def sample_ini_file(tmpdir):
    _path = tmpdir.join('sample_config.ini')
    with open(_path, 'w') as samp_conf_file:
        samp_conf_file.write(SAMPLE_INI_CONTENT)
    yield _path


@pytest.fixture
def ini_config(sample_ini_file):
    _out_cfg = GidAttConfigIni(sample_ini_file)
    _out_cfg.load()
    yield _out_cfg



import pytest
from gidconfig.experimental.experimental_strategy import ConfigStrategos
from gidconfig.experimental.experimental_ini import GidAttConfigIni


def test_is_singleton_ini(sample_ini_file):
    _first = ConfigStrategos.borrow_config(sample_ini_file)
    _second = ConfigStrategos.borrow_config(sample_ini_file)
    assert _first is _second


def test_correct_extension(sample_ini_file):
    _cfg_ini = ConfigStrategos._create_new_config_instance(sample_ini_file)
    assert isinstance(_cfg_ini, GidAttConfigIni)



from gidconfig.standard.classes import ConfigHandler, Get
import pytest


def test_getlist(ini_config):
    assert ini_config.getlist('two', 'list_str_exp') == ["first", "second", "third"]
    assert ini_config.getlist('two', 'list_int_exp') == ["1", "2", "3"]
    assert ini_config.getlist('two', 'list_bool_exp') == ["yes", "no", "True", "false", "off", "on"]
    assert ini_config.getlist('two', 'list_mixed_exp') == ["big", "12", "small", "off", "true", "69"]
    assert ini_config.getlist('two', 'multiword_string_list') == ["this is a", "test of a sentence", "that is split inside a list"]
    assert ini_config.getlist('three', 'list_bad_format') == ["first", "second", "third", "fourth"]

    assert ini_config.getlist('two', 'list_str_exp', as_set=True) == {"first", "second", "third"}

    assert ini_config.getlist('one', 'string_exp') == ['testerstr']



import pytest
from gidconfig.standard.classes import ConfigHandler, Get

import os

SAMPLE_INI_CONTENT = """[one]
string_exp = testerstr
int_exp = 6
bool_exp = yes
multiword_string = this is a test of a sentence
multiword_comma_string = this is a test of a sentence with an, comma in it

[two]
list_str_exp = first, second, third
list_int_exp = 1, 2, 3
list_bool_exp = yes, no, True, false, off, on
list_mixed_exp = big, 12, small, off, true, 69
multiword_string_list = this is a, test of a sentence, that is split inside a list

[three]
list_bad_format = first,second, third,     fourth

"""


@pytest.fixture
def sample_ini_file(tmpdir):
    _path = tmpdir.join('sample_config.ini')
    with open(_path, 'w') as samp_conf_file:
        samp_conf_file.write(SAMPLE_INI_CONTENT)
    yield _path


@pytest.fixture
def ini_config(sample_ini_file):
    _out_cfg = ConfigHandler(sample_ini_file)
    yield _out_cfg
